function [ cropCoords ] = CropBoxFromMask(mask)
    % Argument Check
    if nargin < 1
        theException = MException('CropBoxFromMask:BadArgument', ...
            'No input argument specified.' ...
            );
        throw(theException)
    end
    
    % Is the provided argument a 1-channel mask?
    if size(mask,3) ~= 1
        theException = MException('CropBoxFromMask:BadMask', ...
            'The image provided as an argumenbt is not a 1-channel mask.' ...
            );
        throw(theException)
    end
    
    % For each row of the mask...
    firstRowIndex = 0;
    lastRowIndex = 0;
    for index = 1:size(mask,1)
        thisRow = mask(index,:);
        sumRow = sum(thisRow);
        if sumRow ~= 0
            if firstRowIndex == 0
                firstRowIndex = index;
            end
            lastRowIndex = index;
        end
    end
    
    % For each column of the mask...
    firstColIndex = 0;
    lastColIndex = 0;
    for index = 1:size(mask,2)
        thisCol = mask(:,index);
        sumCol = sum(thisCol);
        if sumCol ~= 0
            if firstColIndex == 0
                firstColIndex = index;
            end
            lastColIndex = index;
        end
    end
    
    cropCoords = [firstColIndex firstRowIndex (lastColIndex - firstColIndex) (lastRowIndex - firstRowIndex)];
end
