#query.py

#Module to resolve a query
import db
import preprocess
import shelve
from Numeric import *
import string
import math

#create a 6x3 matrix for storing weights
#actual weights are read from external file
w = reshape(zeros(6*3, Float), (6,3))

def loadweights(wfile):
    """ loads weights into w from wfile """
    try:
        f = open(wfile)
    except IOError:
        raise IOError, 'Cannot open weights-file'

    lines = f.readlines()
    for i,l in zip(range(0,6),lines):
        w[i] = array([float(x) for x in string.split(l)], Float) #read and store weights

#call loadweights to load the default weights at module import time
loadweights('weights_paint.txt')

def bin(i,j):
    """ function to map coefficients into small number of bins

    it maps i,j to one of the 6 weights
    this functiion here is the one originally used by the authors of the paper
    since weights used are also from the paper the same bin function is used.
    """
    return min(max(i,j),5)


def ScoreQuery(query):
    """ generate scores for query image using m wavelet coefficients

    query is a path to uery image file
    returns a list of image scores
    number of wavelet coefficients is redefined in the preprocess model
    """
    qmat = preprocess.ProcessImage(query)
    if(qmat == None):
        raise AssertionError, 'Cannot preprocess query image'

    scores = zeros(db.size(), Float) #initialize scores for each image

    #open databases for reading
    try:
        idb = shelve.open(db.idbpath,flag='r')
    except:
        raise AssertionError, 'Cannot open image database'


    for c in range(0,3): #for each color channel
        #open searcharrays for this color channel
        # print 'color channel: ', c
        try:
            sa_plus = shelve.open(db.sadbpaths[c][0],flag='r')
            sa_minus = shelve.open(db.sadbpaths[c][1],flag='r')
        except:
            raise IOError, 'Cannot open search array(s) for colorplane ' + str(c)
        
        for iid in idb.keys(): #for each image
            scores[int(iid)] += w[0,c] * math.fabs(qmat[c,0,0] - idb[iid][c+1])

        #set this color channel's avg color val to 0 so that it does not take
        #part in further scoring of images
        qmat[c,0,0] = 0

        for row in range(0,len(qmat[c])):
            indices = nonzero(qmat[c,row])
            for col in indices: #for each non-zero coefficient
                rowcolid = str(row).zfill(3) + str(col).zfill(3)
                #print rowcolid,
                imglist = []
                if(qmat[c,row,col] > 0): #positive search array
                    try:
                        iList = sa_plus[rowcolid]
                        imglist.extend(iList)
                    except KeyError:
                        continue
                else: #negative search array
                    try:
                        iList = sa_minus[rowcolid]
                        imglist.extend(iList)
                    except KeyError:
                        continue

                #print imglist
                for iid in imglist: #for each image in the matched im list
                    scores[int(iid)] -= w[bin(row,col),c] #update the scores

        sa_plus.close()
        sa_minus.close()

    idb.close()
    return scores

